from __future__ import annotations

from typing import Any, Dict, Optional, Union


def pre_parse(
    event: Union[Dict[str, Any], bytes, str], context: Optional[Dict[str, Any]]
) -> Union[Dict[str, Any], bytes, str]:
    """Default no-op pre-parse hook.

    Called after receiving the raw event/context but before input parsing/validation.
    Should return the (possibly modified) event.
    """

    return event


def pre_execute(
    request: Any,
    event: Union[Dict[str, Any], bytes, str],
    context: Optional[Dict[str, Any]],
) -> Any:
    """Default no-op pre-execute hook.

    Called with the Request, original event and context before executing the handler.
    Should return the (possibly modified) Request.
    """

    return request


def pos_execute(response: Any, request: Any) -> Any:
    """Default no-op post-execute hook.

    Called with the handler raw response and the Request,
    before output schema serialization.
    Should return the (possibly modified) response.
    """

    return response


def pre_response(body: Any) -> Any:
    """Default no-op pre-response hook.

    Called with the response body at the end
    of try/except block, just before returning.
    Should return the (possibly modified) body.
    """

    return body



import importlib
import os
import tomllib
from dataclasses import dataclass
from typing import Any, Callable, Dict

HookCallable = Callable[..., Any]


def get_toml_configuration() -> Dict[str, Any]:
    """
    Reads the pyproject.toml file and returns its content as a dictionary.

    Returns:
        A dictionary containing the contents of the pyproject.toml file.
        If the file does not exist or cannot be read, an empty dictionary is returned.
    """
    try:
        with open("pyproject.toml", "rb") as f:
            all_data = tomllib.load(f)
            return all_data.get("tool", {}).get("lamina", {})
    except (FileNotFoundError, tomllib.TOMLDecodeError):
        return {}


@dataclass
class LaminaSettings:
    settings: Dict[str, Any]

    def _get_setting(self, name: str, default: Any = None) -> Any:
        full_name = f"LAMINA_{name.upper()}"
        # First, check environment variables
        value = os.getenv(full_name, None)
        if not value:
            # Then check the settings dictionary
            value = self.settings.get(name, default)

        module_path = (
            ".".join(value.split(".")[:-1]) if ":" not in value else value.split(":")[0]
        )
        func_name = value.split(".")[-1] if ":" not in value else value.split(":")[1]
        try:
            module = importlib.import_module(module_path)
            value = getattr(module, func_name)
        except (ImportError, AttributeError) as error:
            raise ImportError(
                f"Could not import '{value}' for setting '{full_name}'"
            ) from error

        return value

    @property
    def LAMINA_PRE_PARSE_CALLBACK(self) -> HookCallable:
        return self._get_setting("pre_parse_callback", default="lamina.hooks.pre_parse")

    @property
    def LAMINA_PRE_EXECUTE_CALLBACK(self) -> HookCallable:
        return self._get_setting(
            "pre_execute_callback", default="lamina.hooks.pre_execute"
        )

    @property
    def LAMINA_POS_EXECUTE_CALLBACK(self) -> HookCallable:
        return self._get_setting(
            "pos_execute_callback", default="lamina.hooks.pos_execute"
        )

    @property
    def LAMINA_PRE_RESPONSE_CALLBACK(self) -> HookCallable:
        return self._get_setting(
            "pre_response_callback", default="lamina.hooks.pre_response"
        )


# Create a single instance of the settings class
_lamina_settings = LaminaSettings(get_toml_configuration())


def __getattr__(name: str) -> Any:
    """
    Implement PEP 562 __getattr__ to lazily load settings.

    This function is called when an attribute is not found in the module's
    global namespace. It delegates to the _lamina_settings instance.
    """
    return getattr(_lamina_settings, name)



import asyncio
import functools
import inspect
import json
from dataclasses import dataclass
from typing import Any, Callable, Dict, Optional, Type, TypedDict, Union

import magic
from loguru import logger
from pydantic import BaseModel, RootModel, ValidationError

from lamina import conf
from lamina.helpers import DecimalEncoder


@dataclass
class Request:
    """Request object passed to decorated handlers.

    Attributes:
        data: Parsed body or model instance according to schema_in and flags.
        event: Original AWS Lambda event.
        context: Lambda context object.
    """

    data: Union[BaseModel, str]
    event: Union[Dict[str, Any], bytes, str]
    context: Optional[Dict[str, Any]]
    pre_execute_result: Optional[Any] = None


class ResponseDict(TypedDict):
    statusCode: int
    headers: Dict[str, str]
    body: str


def lamina(
    schema_in: Optional[Type[BaseModel] | Type[RootModel]] = None,
    schema_out: Optional[Type[BaseModel] | Type[RootModel]] = None,
    content_type: str | None = None,
    step_functions: bool = False,
) -> Callable[[Callable[..., Any]], Callable[..., ResponseDict]]:
    def decorator(f: Callable[..., Any]) -> Callable[..., ResponseDict]:
        @functools.wraps(f)
        def wrapper(
            event: Dict[str, Any] | bytes | str,
            context: Optional[Dict[str, Any]],
            *args: Any,
            **kwargs: Any,
        ) -> ResponseDict:
            if f.__doc__:
                title = f.__doc__.split("\n")[0].strip()
            else:
                # event may not be a dict; guard get
                path = event.get("path") if isinstance(event, dict) else "unknown"
                title = f"{f.__name__} for path {path}"
            logger.info(f"******* {title.upper()} *******")
            logger.debug(event)

            magic_content_type = "application/json"

            try:
                # Run pre-parse hook (may adjust event)
                pre_parse_hook = conf.LAMINA_PRE_PARSE_CALLBACK
                if inspect.iscoroutinefunction(pre_parse_hook):
                    event = asyncio.run(pre_parse_hook(event, context))
                else:
                    event = pre_parse_hook(event, context)

                # Parse input (after possible pre-parse modification)
                if schema_in is None:
                    data = (
                        event["body"]
                        if (isinstance(event, dict) and not step_functions)
                        else event
                    )
                else:
                    request_body = (
                        json.loads(event["body"])
                        if (isinstance(event, dict) and not step_functions)
                        else event
                    )
                    data = schema_in(**request_body)

                # Build initial Request and run pre-execute hook
                request = Request(
                    data=data,
                    event=event,
                    context=context,
                )
                pre_execute_hook = conf.LAMINA_PRE_EXECUTE_CALLBACK
                if inspect.iscoroutinefunction(pre_execute_hook):
                    request = asyncio.run(pre_execute_hook(request, event, context))
                else:
                    request = pre_execute_hook(request, event, context)

                status_code = 200

                headers: Dict[str, str] = {}

                # check if function is a coroutine
                if inspect.iscoroutinefunction(f):
                    response: Any = asyncio.run(f(request))
                else:
                    response = f(request)

                # Execute post-execution hook on raw response (before schema_out)
                pos_execute_hook = conf.LAMINA_POS_EXECUTE_CALLBACK
                if inspect.iscoroutinefunction(pos_execute_hook):
                    response = asyncio.run(pos_execute_hook(response, request))
                else:
                    response = pos_execute_hook(response, request)

                if isinstance(response, tuple):
                    status_code = response[1]
                    if len(response) == 3:
                        headers = response[2]
                    response = response[0]

                try:
                    body: str | Any = response
                    if body:
                        if schema_out:
                            if issubclass(schema_out, RootModel):
                                root = schema_out(response).root
                                if root is not None:
                                    body = schema_out(response).model_dump_json(by_alias=True) if not isinstance(root, str) else root
                            else:
                                body = schema_out(**response).model_dump_json(
                                    by_alias=True
                                )
                        body = json.dumps(body, cls=DecimalEncoder) if not isinstance(body,str) else body
                    magic_content_type = magic.from_buffer(body, mime=True) if body else "text/html"
                except Exception as e:
                    # This is an Internal Server Error
                    logger.error(f"Error when attempt to serialize response: {e}")
                    status_code = 500
                    body = json.dumps(
                        [
                            {
                                "field": (
                                    schema_out.__name__ if schema_out else "DumpJson"
                                ),
                                "message": str(e),
                            }
                        ],
                        cls=DecimalEncoder,
                    )

                full_headers: Dict[str, str] = {
                    "Content-Type": content_type
                    or f"{magic_content_type}; charset=utf-8",
                }
                if headers:
                    full_headers.update(headers)

                # Run pre-response hook just before returning
                pre_response_hook = conf.LAMINA_PRE_RESPONSE_CALLBACK
                if inspect.iscoroutinefunction(pre_response_hook):
                    body = asyncio.run(pre_response_hook(body))  # type: ignore[misc]
                else:
                    body = pre_response_hook(body)

                return {
                    "statusCode": status_code,
                    "headers": full_headers,
                    "body": body,  # type: ignore[return-value]
                }
            except ValidationError as e:
                messages = [
                    {
                        "field": (
                            error["loc"][0] if error.get("loc") else "ModelValidation"
                        ),
                        "message": error["msg"],
                    }
                    for error in e.errors()
                ]
                logger.error(messages)
                body = json.dumps(messages)
                return {
                    "statusCode": 400,
                    "body": body,
                    "headers": {
                        "Content-Type": "application/json; charset=utf-8",
                    },
                }
            except (ValueError, TypeError) as e:
                message = f"Error when attempt to read received event: {event}."
                logger.error(str(e))
                body = json.dumps(message)
                return {
                    "statusCode": 400,
                    "body": body,
                    "headers": {
                        "Content-Type": "application/json; charset=utf-8",
                    },
                }
            except Exception as e:
                logger.exception(e)
                body = json.dumps({"error_message": str(e)})
                return {
                    "statusCode": 500,
                    "body": body,
                    "headers": {
                        "Content-Type": "application/json; charset=utf-8",
                    },
                }

        wrapper.schema_in = schema_in
        wrapper.schema_out = schema_out
        wrapper.content_type = content_type

        return wrapper

    return decorator
